/* dbug.c - 80x86 code debugger main file */

#include "general.h"

#ifdef INC_UNASM
# include "disasm.h"
#endif /*INC_UNASM*/

#ifdef INC_DOSDIR
# include "dosdir.h"
#endif /*INC_DOSDIR*/


#ifdef INC_LOADCORE
# define REG_BASE        0xfb4 /* based on carefull dissection.. */
#endif /*INC_LOADCORE*/

#define DUMP_STD_SZ     0x80

#ifdef INC_UNASM
# define UNASM_STD_SZ    0x20
#endif /*INC_UNASM*/

void (*dir_function)(void);
void (*recurse_function)(position);


/* NOTE: not now, but eventually want to incorporate assembler into this
 */
/* COMMANDS:
 *  (opsize) report or set default size to 2 or 4 bytes
 *  (filesize) sz
 *  (u)nassemble
 *  (d)ump
 *  (l)oad
 *  (r)regs
 *  (h)ex add and sub
 *  loadcore
 *  (s)earch
 *  (m)ove
 *  (q)uit
 *  (?) (help) - list of commands
 *  (save) filename
 *  (multisearch) specfile resultfile
 *  (writeable) flag_val
 *  (read) posn filename
 *  (writeback)
 *  (e)nter x y z
 *  (debug)
 *  (setdos) cluster_start_offset cluster_size\n");
 *  (dir) start end {you must set??? first}\n");
 *  (recurse_save) dir_posn cpio_file_name {save dirs recursively}\n");
 */


   /* error functions */

void error(char *str)
{ fprintf(stderr,"error: %s\n",str);
}

void fatalerror(char *str)
{ fprintf(stderr,"fatal error: %s\n",str);
 exit(-1);
}


    /* ascii map (binary-->printable ascii approximation) */

unsigned char ascii_map[0x100];

static void init_ascii_map(void)
{ int a;

    /* this will print all 256 except control chars literally: */
  for(a=0;a<' ';a++) ascii_map[a]='.'; /* PORT: assumes ASCII */
  for(;a<256;a++)    ascii_map[a]=a;
#ifdef ASCII_CLEAN
  ascii_map[127]='.'; /* del */
  for(a=128;a<256;a++) ascii_map[a]='.';
#endif /*ASCII_CLEAN*/
}


     /* dissassembler interface functions */

#ifdef INC_UNASM
static position dis_scan,dis_end;
#endif /*INC_UNASM*/

static struct registers      /* organized to work with defns in <sys/reg.h> */
{ int gs,fs,es,ds;       /* 00-03 */
  int edi,esi,ebp,kesp;  /* 04-07 */
  int ebx,edx,ecx,eax;   /* 08-0b */
  int trapno,err,eip,cs; /* 0c-0f */
  int efl,uesp,ss;       /* 10-12 */
} regs;

static int operand_size=4; /* future: may be part of assembler */

#ifdef INC_UNASM
  /* functions to interface with dissassembler */
static int dis_out(outstr) char *outstr;
{ printf("%s\n",outstr);
}

static int dis_stop()
{ return(dis_scan>=dis_end);
}

static int dis_address(str) char *str;
{ sprintf(str,"%08x",dis_scan);
}

static int dis_offset()
{ return dis_scan;
}

static int dis_by()
{ if(dis_scan>=infosz) return 0;
  else return bf_byte(dis_scan++);
}

static int dis_wo()
{ if(dis_scan>infosz-sizeof(word)) return 0;
  else
  { int a;

    a=dis_scan;
    dis_scan+=sizeof(word);
   return bf_word(a);
  }
}

static int dis_dw()
{ if(dis_scan>infosz-sizeof(dword)) return 0;
  else
  { int a=dis_scan;
    dis_scan+=sizeof(dword);
   return bf_dword(a);
  }
}
#endif /*INC_UNASM*/

  /* file i/o functions */
static int file_load(str) char *str;
{ bf_init(str,1);
}

#ifdef INC_LOADCORE
static int file_core_load(str) char *str;
{ FILE *file;
  user_t user;

  file=fopen(str,"r");
  if(!file) fprintf(stderr,"Cannot open file \"%s\" for reading.\n",str);
  else
  { fread(&user,sizeof(user),1,file);
    printf(
      "sizeof(user) %#x user.u_ar0 %#x u_tsize %#x u_dsize %#x u_ssize %#x\n"
      ,sizeof(user),user.u_ar0,user.u_tsize,user.u_dsize,user.u_ssize);
    printf("offset(user.u_ar0) %#x\n",(char *)&user.u_ar0-(char *)&user);
    fclose(file);
    file_load(str);
    memcpy(&regs,info+REG_BASE,sizeof(regs));
  }
}
#endif /*INC_LOADCORE*/


      /* search functions */

struct search_thing
{ byte *srch;   /* string to search for */
  int len;      /* length of string */
  int off;      /* key char to look for (optimization) */
  struct search_thing *next; /* next thing with same char */
};

#define SEQ_SIZE 128

byte *get_byte_seq(int *outlen
         ,int (*getch)(),position (*getnum)(),int (*getstr)())
   /* get a byte sequence from input functions.  Sequence contains a
    * combination of byte values and literal strings: ["This is a test" a d 0]
    * warning: reuses same output buffer each time called. */
{ static byte *buf;
  static int buf_al;
  int len,ch;
  int a;

  if(!buf)
  { buf_al=SEQ_SIZE;
    buf=malloc(buf_al);
    if(!buf)
    { fatalerror("out of memory in get_byte_seq()");
    }
  }

  len=0;
  while(1)
  { ch=(*getch)();
    if(ch=='\n') break;
    if(ch==EOF) break;
    if(ch=='\"')
    { byte tempstr[MAXSTRLEN+1];

      if(!(*getstr)(tempstr)) return NULL;
      a=strlen(tempstr);
      if(len+a>buf_al)
      { buf_al=len+SEQ_SIZE;
        buf=realloc(buf,buf_al);
        if(!buf) fatalerror("out of memory 2 in get_byte_seq()");
      }
      memcpy(buf+len,tempstr,a); len+=a;
    }
    else if(isxdigit(ch))
    { a=(*getnum)(ch);
      if(buf_al<=len)
      { buf_al=len+SEQ_SIZE;
        buf=realloc(buf,buf_al);
        if(!buf) fatalerror("out of memory 3 in get_byte_seq()");
      }
      buf[len++]=a;
    }
    else if(!isspace(ch))
    { error("syntax error in input.");
     return NULL;
    }
  }
  if(len==0&&ch==EOF) return NULL;
  *outlen=len;
 return buf;
}

#define S_ENDLIST    NULL

static void do_search(FILE *resultfile,struct search_thing *ms,int cnt
       ,position start,position end)
    /* search engine.  Searches for cnt strings simultaneously given in ms.
     * Searches from start to end.
     */
{ int a,b;
  struct search_thing *map[256];
  struct search_thing *str;
  position off,size;
  byte *ptr;

  if(end>infosz) end=infosz;

    /* form key character map: */
  for(a=0;a<256;a++) map[a]=S_ENDLIST;
  for(a=0;a<cnt;a++)
  { b=ms[a].srch[ms[a].off];
    ms[a].next=map[b];
    map[b]=ms+a;
  }

   /* search block:
    * for each char
    *   decide weather we need to check further by map[char]
    *     if so, check everything that matches that char.
    */
  for(off=start;off<end;off+=size)
  { ptr=bf_loadchunk(off,&size);
    if(off+size>end) size=end-off;
    for(a=0;a<size;a++)
    { if((str=map[ptr[a]])!=S_ENDLIST)
      { for(;str!=S_ENDLIST;str=str->next)
        { if(str->off>off+a) continue;
          if(!bf_memcmp(off+a-str->off,str->srch,str->len))
          { fprintf(resultfile,"%x %x\n",str-ms,off+a-str->off);
          }
        }
        ptr=bf_loadchunk(off,&size);
      }
    }
  }
}

static FILE *srch_file;

static int srch_getch(void)
{ return getc(srch_file);
}

static position srch_getnum(int ch)
{ position a;

  ungetc(ch,srch_file);
  fscanf(srch_file,"%x",&a);
 return a;
}

static int srch_getstr(byte *string)
{ return fscanf(srch_file,"%[^\"]%*c",string)>0;
}

void multisearch(char *specs,char *results)
   /* searches for stuff given in file specs; puts search results
    * in file results */
{ FILE *file;
  byte *buf;
  int len;
  int a;
  struct search_thing *ms;

  buf=NULL;

     /* get specifications for search: */

  srch_file=file=fopen(specs,"r");
  if(!file)
  { fprintf(stderr,"Can't open spec file %s for input\n",specs);
   return;
  }

  ms=(struct search_thing *)malloc(sizeof(*ms));
  a=0;

  while(fscanf(file,"%d",&ms[a].off)==1)
  { buf=get_byte_seq(&len,srch_getch,srch_getnum,srch_getstr);

    if(!buf) break;
    if(len==0) continue;

    ms[a].srch=Mcpy(buf,len);
    ms[a].len=len;
    if(ms[a].off<0||ms[a].off>=len)
    { error("Bad key char offset");
     return;
    }
    a++;
    ms=(struct search_thing *)realloc(ms,(a+1)*sizeof(*ms));
  }
  fclose(file);

     /* do search: */

  if(!*results) file=stdout;
  else file=fopen(results,"w");
  if(!file)
  { fprintf(stderr,"Can't open results file");
  }
  else
  { do_search(file,ms,a,0,infosz);
    if(file!=stdout) fclose(file);
  }

    /* clean up: */
  free(ms);
  Mfreeall();
}

      /* hexd functions */

static int dump_scan;

static int dump_getby()
{ if(dump_scan>=infosz) return 0;
  else return bf_byte(dump_scan++);
}

static void dump_getchunk(chunk) char chunk[16];
{ byte *ptr;
  int size;

  ptr=bf_16(dump_scan,&size);
  if(size<16) memcpy(chunk,ptr,size);
  else memcpy(chunk,ptr,16);
  dump_scan+=16;
}

void dump_func(f_off,len) position f_off,len;
{ int l_sz;       /* number of chars. in line */
  int startposn;
  unsigned char buf[16];
  /* PORT: assume this really is unsigned when converted to int */
  int ch,a;
  static char ch1[256],ch2[256]; /* MSD=ch1,  LSD=ch2 */
  static char ch_space[16];      /* spaceing after value */
  static init_ch1; /* are the above arrays initialized? */

  if(!init_ch1) /* initialize printing optimization chars. */
  { init_ch1=1;
    for(a=0;a<256;a++)
    { ch=(a>>4)&0xf;
      ch1[a]=ch<10?ch+'0':ch+('a'-10);
      ch=a&0xf;
      ch2[a]=ch<10?ch+'0':ch+('a'-10);
    }
    for(a=0;a<16;a++) ch_space[a]=(a==7?'-':' ');
  }

  startposn=f_off%(unsigned)16;
  do
  { if(!startposn&&startposn+len>=16)               /* get 16 bytes */
    { dump_getchunk(buf);
      len-=16;
      l_sz=16;
    }
    else for(l_sz=startposn;l_sz<16&&len>0;l_sz++)
      { ch=dump_getby();
        len--;
        buf[l_sz]=ch;
      }
    if(l_sz==startposn) break;

    printf("%08x%*s",(f_off/16*16),2+(startposn*3),""); /* print offset */

    for(a=startposn;a<l_sz;a++)                   /* output in hex */
    { ch=buf[a];
      putchar(ch1[ch]); putchar(ch2[ch]); putchar(ch_space[a]);
        /* ABOVE DOES: printf("%02x%s",buf[a],a==7?"-":" "); ONLY FASTER */

      buf[a]=ascii_map[ch];                 /* re-map data for char. out */
    }

      /* pad out unused hex stuff, and output as chars. */
    printf("%*s%.*s\n",(16-l_sz)*3+startposn+2,""
         ,l_sz-startposn,buf+startposn);

    f_off+=l_sz-startposn;
    startposn=0;
  } while(l_sz==16);
}


      /* auxiliary support functions for main */

static int getvalue(str,end) char *str; int *end;
{ int a,out;
  a=sscanf(str,"%x%n",&out,end);
 return a<1?-1:out;
}

static position getposn(str,end) char *str; int *end;
{ return getvalue(str,end);
}

    /* getdata() [get_byte_seq() from standard input] */

static byte *gd_str;

static int gd_getch(void)
{ if(!*gd_str) return EOF;
 return *(gd_str++);
}

static position gd_getnum(int ch)
{ int len= -1;
  position out;

  if(sscanf(gd_str-1,"%x%n",&out,&len)!=1||len<=0)
      fatalerror("internal lexer error");
  gd_str+=len-1;
 return out;
}

static int gd_getstr(byte *string)
{ char junkchar;
  int len= -1;

  if(sscanf(gd_str,"%[^\"]%c%n",string,&junkchar,&len)!=2||len<=0)
  { error("bad string");
    *string=0;
   return 0;
  }
  gd_str+=len;
 return 1;
}

static byte *getdata(byte *str,int *outlen)
{ gd_str=str;
 return get_byte_seq(outlen,gd_getch,gd_getnum,gd_getstr);
}


           /* ******* main token input section ******** */

static char l_commandline[MAXSTRLEN+1];
static int l_commandposn;

position l_getposn(void)
{ position data;
  int len;

  if(sscanf(l_commandline+l_commandposn,"%x%n",&data,&len)!=1)
  { data= -1;
  }
  else l_commandposn+=len;
 return data;
}

int l_getval(void)
{ return l_getposn();
}

int l_getstr(char *str)
{ int len;

  *str=0;
  if(sscanf(l_commandline+l_commandposn,"%s%n",str,&len)!=1)
  {return 0;
  }
  l_commandposn+=len;
 return 1;
}

byte *l_getdata(int *len)
{ return getdata(l_commandline+l_commandposn,len);
}

void get_range(position *start,position *end,int def_sz)
{ position pn;

  pn=l_getposn();
  if(pn!=-1)
  { (*start)=pn;
    (*end)=l_getposn();
  }
  else (*end)= -1;
  if((*start)>=infosz) (*start)=0;
  if((*end)==-1) (*end)=(*start)+def_sz;
  if((*end)>infosz) (*end)=infosz;
}


        /* ************* command implementations *********** */

void cmd_filesize(void)
   /* change file size */
{ position pn;

  pn=l_getposn();
  bf_changesize(pn);
}

void cmd_opsize(void)
   /* change default operand size */
{ int a;

  a=l_getval();
  if(a==2||a==4) operand_size=a;
  printf("size is now %d bytes\n",operand_size);
}

#ifdef INC_UNASM
void cmd_u(void)
    /* unassemble */
{ int a;

  get_range(&dis_scan,&dis_end,UNASM_STD_SZ);

  a=dissassemble(dis_out,dis_address,dis_offset,dis_stop,dis_by
        ,dis_wo,dis_dw,operand_size,1);
  if(a!=STATUS_OK)
  { fprintf(stderr
      ,"disassembler unit reported an abnormal termination status number %d\n"
       ,a);
  }
}
#endif /*INC_UNASM*/

void cmd_d(void)
   /* hex dump */
{ position pn;

  get_range(&dump_scan,&pn,DUMP_STD_SZ);
  pn-=dump_scan;
  dump_func(dump_scan,pn);
}

void cmd_l(void)
    /* load new file */
{ char tempstr[MAXSTRLEN+1];

  l_getstr(tempstr);
  if(!*tempstr)
  { fprintf(stderr,"\"%s\": missing file name\n",l_commandline);
  }
  else
  { bf_write_back_all();
    file_load(tempstr);
  }
}

void cmd_r(void)
   /* print registers and others */
{ static char *(writeable[])
     ={ "readonly","nonpersistent writeable","persistent writeable" };

#ifdef INC_LOADCORE
  printf("eax %-8x   ecx %-8x   edx %-8x   ebx %-8x\n"
     ,regs.eax,regs.ecx,regs.edx,regs.ebx);
  printf("uesp %-8x  ebp %-8x   esi %-8x   edi %-8x\n"
     ,regs.uesp,regs.ebp,regs.esi,regs.edi);
  printf("kesp %-8x  err %-8x   eip %-8x   efl %-8x\n"
     ,regs.kesp,regs.err,regs.eip,regs.efl);
  printf("cs %-8x    ds %-8x    es %-8x    fs %-8x\n"
     ,regs.cs,regs.ds,regs.es,regs.fs);
  printf("gs %-8x    ss %-8x    trapno %-8x\n"
     ,regs.gs,regs.ss,regs.trapno);
#endif /*INC_LOADCORE*/
  printf("File \"%s\" is%s loaded.\n",file_name,bf_loaded?"":" not");
  printf("logical file size %-8x   physical size %-8x\n",infosz,real_infosz);
  printf("writeable flag %d (%s)\n",bf_writeable,writeable[bf_writeable]);
  printf("default operand size %d\n",operand_size);
#ifdef INC_DOSDIR
  printf("dos start of clusters %x   cluster size %x\n",dos_cstart,dos_csize);
#endif /*INC_DOSDIR*/
}

void cmd_h(void)
   /* hexidecimal arithmetic */
{ int val1,val2;

  val1=l_getval();
  val2=l_getval();
  if(val2==-1) /* if val1 is bad, this will be bad too. */
  { fprintf(stderr,"\"%s\": Bad values.\n",l_commandline);
   return;
  }
  printf("+ %-8x   - %-8x   R %-8x   & %-8x   | %-8x   ^ %x\n"
    ,val1+val2,val1-val2,val2-val1,val1&val2,val1|val2,val1^val2);
  printf("* %-8x   / %-8x   %% %-8x  >> %-8x  << %-8x   ~ %x\n"
    ,val1*val2,val1/val2,val1%val2,val1>>val2,val1<<val2,~val1);
}

void cmd_s(void)
   /* simple search */
{ position search_scan,search_end;
  struct search_thing ms;
  int a;

  search_scan=l_getposn();
  if(search_scan<infosz)
  { search_end=l_getposn();
  }
  else search_end= -1;
  if(search_end<search_scan||search_end==-1)
  { fprintf(stderr,"\"%s\": Bad range.\n",l_commandline);
   return;
  }
  ms.srch=l_getdata(&a);
  if(!ms.srch) return;
  ms.off=0;
  ms.len=a;
  do_search(stdout,&ms,1,search_scan,search_end);
}

#ifdef INC_LOADCORE
void cmd_loadcore(void)
{ sscanf(commandline+argumentbase," %s",tempstr);
  file_core_load(tempstr);
}
#endif /*INC_LOADCORE */

void cmd_m(void)
    /* move chunk of file */
{ position start,end,backend;
  position new;
  int size,nsize;
  byte *sptr,*dptr;

  start=end= -1;
  get_range(&start,&end,0);
  new=l_getposn();

  if(new==-1||start>end||end>infosz)
  { fprintf(stderr,"Bad usage; see help\n");
   return;
  }

      /* fix size: */
  if(end-start+new>infosz)
  { if(!bf_changesize(end-start+new)) end=start+infosz-new;
  }

  if(new>start)
  {   /* start from end */
    new+=end-start; /* new now points to end of dest. */
    while(start<end)
    { sptr=bf_loadchunk_back(end,&size);

      dptr=bf_loadchunk_back(new,&nsize);
      bf_modifiing(new-nsize);

      if(nsize<size)
      { sptr+=(size-nsize);
        size=nsize;
      }
      else if(nsize>size)
      { dptr+=(nsize-size);
      }
      if(size>end-start)
      { nsize=size-(end-start);
        size=end-start;
        dptr+=nsize;
        sptr+=nsize;
      }
      memmove(dptr,sptr,size);
      new-=size;
      end-=size;
    }
  }
  else if(new<start)
  {   /* start from start */
    while(start<end)
    { sptr=bf_loadchunk(start,&size);

      dptr=bf_loadchunk(new,&nsize);
      bf_modifiing(new);

      if(nsize<size) size=nsize;
      if(start+size>end) size=end-start;
      
      memmove(dptr,sptr,size);
      new+=size;
      start+=size;
    }
  }
}

void cmd_q(void)
   /* quit */
{ bf_write_back_all();
 exit(-1);
}

void cmd_e(void)
    /* enter new data */
{ position place,size;
  int a;
  byte *ptr;
  byte *data;

  place=l_getposn();
  if(place==-1)
  { fprintf(stderr,"\"%s\": illegal position\n",l_commandline);
   return;
  }
  data=l_getdata(&a);
  if(!data)
  { error("no data to enter");
   return;
  }
  while(a>0) /* not optimized */
  { if(place>=infosz&&!bf_changesize(place+1)) break;
    ptr=bf_loadchunk(place,&size);
    if(!size)
    { error("unexpected at EOF");
     return;
    }
    bf_modifiing(place++);
    *ptr=*data++; a--;
  }
}

void cmd_help(void)
    /* print command summaries */
{ printf(" (opsize) set default size to 2 or 4 bytes\n");
#ifdef INC_UNASM
  printf(" (u)nassemble\n");
#endif /*INC_UNASM*/
  printf(" (filesize) sz\n (d)ump\n (l)oad\n (r)regs (and other)\n");
  printf(" (save) file start end\n (e)nter start y y y\n");
  printf(" (h)ex add and sub\n");
#ifdef INC_LOADCORE
  printf(" (loadcore)\n");
#endif /*INC_LOADCORE*/
  printf(" (s)earch\n (m)ove start end new\n");
  printf(" (q)uit\n (?) (help) - list of commands\n");
  printf(" (multisearch) specfile resultfile\n");
  printf(" (writeable) flag_val [0(ro) 1(rw) 2(rw survives l)]\n");
  printf(" (read) posn filename\n (writeback)\n");
  printf(" (debug)  - not always implemented\n");
#ifdef INC_DOSDIR
  printf(" (setdos) cluster_start_offset cluster_size\n");
#endif /*INC_DOSDIR*/
  printf(" (dir) start end {you must set??? first}\n");
  printf(" (recurse_save) dir_posn cpio_file_name {save dirs recursively}\n");
}

void cmd_writeable(void)
    /* adjust writeable flag */
{ int a;

  a=l_getval();
  bf_set_writeable(a);
}

void do_file_write(position start,position end,FILE *file)
{ byte *ptr;
  int size;
  position pn;

  for(pn=start;pn<end;pn+=size)
  { ptr=bf_loadchunk(pn,&size);
    if(pn+size>end) size=end-pn;
    if(size==0) return;
    if(size<0||fwrite(ptr,1,size,file)!=size)
    { fprintf(stderr,"Error writing file.\n");
     return;
    }
  }
}

void cmd_save(void)
   /* save some or all of file */
{ FILE *file;
  position start,end;
  char tempstr[MAXSTRLEN+1];

  l_getstr(tempstr);
  if(!*tempstr)
  { error("Need string");
   return;
  }

  start=0;
  get_range(&start,&end,infosz);

  file=fopen(tempstr,"w");
  if(!file)
  { fprintf(stderr,"Cannot open file \"%s\" for writing.\n",tempstr);
   return;
  }
  do_file_write(start,end,file);
  fclose(file);
}

void cmd_debug(void)
    /* print self-diagnostic information */
{ bf_debug();
}

void cmd_multisearch(void)
    /* search for stuff in spefile; put result in resultfile */
{ char specfile[MAXSTRLEN+1],resultfile[MAXSTRLEN+1];

  l_getstr(specfile);
  l_getstr(resultfile);
  if(!*specfile)
  { error("need specifications file name");
  }
  else multisearch(specfile,resultfile);
}

#define READ_BUF_SZ (MAXSTRLEN+1+1024)
void cmd_read(void)
   /* read a new file over portion of current file */
{ char tempstr[READ_BUF_SZ];
  position pn;
  FILE *file;
  int size,rsize;
  byte *ptr;

  pn=l_getposn();
  l_getstr(tempstr);
  if(pn==-1||!*tempstr)
  { fprintf(stderr,"\"%s\": bad usage\n",l_commandline);
  }
  else
  { file=fopen(tempstr,"r");
    if(!file)
    { fprintf(stderr,"Can't open file %s for input\n",tempstr);
     return;
    }
    do
    { if(pn==infosz)
      { size=READ_BUF_SZ;
        rsize=fread(tempstr,1,size,file);
        if(rsize>0)
        { int remain=rsize,cnksize;
          int tsoff=0;

          if(!bf_changesize(pn+rsize))
          { fclose(file);
           return;
          }
          while(remain)
          { ptr=bf_loadchunk(pn,&cnksize);
            bf_modifiing(pn);
            if(cnksize>remain) cnksize=remain;
            memcpy(ptr,tempstr+tsoff,cnksize);
            pn+=cnksize;
            tsoff+=cnksize;
            remain-=cnksize;
          }
        }
      }
      else
      { ptr=bf_loadchunk(pn,&size);
        if(size==0) error("internal; zero size");
        bf_modifiing(pn);
        rsize=fread(ptr,1,size,file);
        pn+=rsize;
      }
    } while(rsize==size);
    fclose(file);
  }
}

void cmd_writeback(void)
   /* write back all changes to file */
{ bf_write_back_all();
}

#ifdef INC_DOSDIR
void cmd_setdos(void)
{ position cstart,csize;

  cstart=l_getposn();
  csize=l_getposn();
  if(cstart!=-1||csize!=-1)
  { dos_cstart=cstart;
    if(csize!=-1) dos_csize=csize;
  }
  printf("cluster start %x   size %x\n",dos_cstart,dos_csize);
  dir_function=do_dos_dir;
  recurse_function=dos_recurse;
}
#endif /*INC_DOSDIR*/

void cmd_dir(void)
{ if(dir_function) (*dir_function)();
  else error("No directory function set");
}

void cmd_recurse_save(void)
{ position start=0,end;

  start=l_getposn();
  if(recurse_function) (*recurse_function)(start);
  else error("No directory function set");
}


      /* ---------- main ---------- */

static struct commands
{ char *name;
  void (*func)();
} cmds[]
 ={ { "opsize",cmd_opsize },
    { "filesize",cmd_filesize },
#ifdef INC_UNASM
    { "u",cmd_u },
#endif /*INC_UNASM*/
    { "d",cmd_d },
    { "l",cmd_l },
    { "r",cmd_r },
    { "h",cmd_h },
    { "s",cmd_s },
#ifdef INC_LOADCORE
    { "loadcore",cmd_loadcore },
#endif /*INC_LOADCORE*/
    { "e",cmd_e },
    { "m",cmd_m },
    { "q",cmd_q },
    { "?",cmd_help },
    { "help",cmd_help },
    { "writeable",cmd_writeable },
    { "save",cmd_save },
    { "multisearch",cmd_multisearch },
    { "read",cmd_read },
    { "writeback",cmd_writeback },
    { "debug",cmd_debug },
#ifdef INC_DOSDIR
    { "setdos",cmd_setdos },
#endif /*INC_DOSDIR*/
    { "dir",cmd_dir },
    { "recurse_save",cmd_recurse_save },
    { "",NULL }
  };

int main(int argc,char *argv[])
{ char tempstr[MAXSTRLEN+1];
  int len;
  int a;
  position pn;
  char *filename;
  int got_file=0;
  int wanted_writeable=0;

  init_ascii_map();

    /* parse command line options */
  for(a=1;a<argc;a++)
  { int stat=0; /* used either for exit status or offset in argv[a] of file */
    int indx;

    if(*argv[a]=='-')
    { if(!strcmp(argv[a],"--help")) goto o_help;
      else if(!strcmp(argv[a],"--readonly"))
      { wanted_writeable=0;
      }
      else if(!strcmp(argv[a],"--writeable"))
      { if(++wanted_writeable>2) wanted_writeable=2;
      }
      else if(!strncmp(argv[a],"---",3))
      { stat=2; /* skip first two chars */
       goto o_isfile;
      }
      else for(indx=1;argv[a][indx];indx++)
      { switch(argv[a][indx])
        { case 'h':
          o_help:
           cmd_help();
           printf("usage: %s [options] [filename]\n",argc?argv[0]:"dbug");
           printf("  options:\n    -h, --help,\n    -r, --readonly,\n");
           printf("    -w, --writeable, [two times for permanent w]\n");
           printf("  file names like \"---file\" interpreted as \"-file\".\n");
          exit(stat);
          case 'r':
          o_readonly:
           wanted_writeable=0;
          break;
          case 'w':
          o_writeable:
           if(++wanted_writeable>2) wanted_writeable=2;
          break;
          default:
           fprintf(stderr,"bad option\n");
           stat= -2; /* error status */
          goto o_help;
        }
      }
    }
    else
    {o_isfile:
      if(got_file)
      { fprintf(stderr,"too many files in command line\n");
        stat= -3;
       goto o_help;
      }
      got_file=1;
      filename=argv[a]+stat;
    }
  }
  if(got_file)
  { file_load(filename);
    bf_set_writeable(wanted_writeable);
  }
  else
  { if(wanted_writeable==1)
    { error("Can't set nonpersistent writeable without a file!");
    }
    bf_init("(None)",0);
  }

    /* parse/interpret interactive commands: */
  while(1)
  { printf("* ");
    if(fgets(l_commandline,MAXSTRLEN,stdin)==NULL)  /* get command and args */
    { fprintf(stderr,"Failed to read anything from stdin. Stopping.\n");
     exit(-1);
    }
    len=strlen(l_commandline);                      /* check termination */
    if(len==1) continue; /* blank command */
    if(!len||l_commandline[len-1]!='\n')
    { fprintf(stderr,"Weird form of input, ignoreing.\n");
     continue;
    }
    l_commandline[len-1]='\0';                    /* junk the newline */
    l_commandposn=0;

       /* get command */
    l_getstr(tempstr);
    if(!*tempstr) continue;

        /* find coresponding function (this isn't optimized) */
    for(a=0;*cmds[a].name;a++)
    { if(!strcmp(tempstr,cmds[a].name))
      { (*cmds[a].func)();
       break;
      }
    }

    if(!*cmds[a].name)
    { fprintf(stderr, "Unknown command \"%s\"\n",l_commandline);
    }
  }
}
