/* dosdir.c - dos directry interpreter */

#include "general.h"

#ifdef INC_DOSDIR

#include "dosdir.h"

position dos_csize,dos_cstart;

#define DOS_NAME_LEN       8
#define DOS_EXT_LEN        3
#define DOS_DIR_ENTRY_SIZE 32
#define DOS_TOTAL_LEN      (DOS_NAME_LEN+DOS_EXT_LEN+1)

struct processed_dos_dir
{ char *name;
  int attrib;
  int cluster;
  position size;
  position posn;
  position start;
  int good_entry;
};

static void do_scan_dir(position *start,position end
     ,void (*interpret)(struct processed_dos_dir *))
{ int a,b;
  struct dos_dir
  { byte name[DOS_NAME_LEN];
    byte ext[DOS_EXT_LEN];
    byte attrib;
    byte misc1[10+2+2];
    byte clust1,clust2;    /* Break these up: not sensitive */
    byte sz1,sz2,sz3,sz4;  /*   to endianness. */
  } *dir;
  position size,cluster;
  char name[DOS_NAME_LEN+DOS_EXT_LEN+5]; /* at least +2 (see sprintf) */
  struct processed_dos_dir pdir;
  position dstart=*start;
  position posn;
  struct dos_dir good_entry; /* used as overflow/copy place for bf_buffer() */
  char map_name[DOS_NAME_LEN],map_ext[DOS_NAME_LEN];

  for(a=0;dstart<end;dstart+=DOS_DIR_ENTRY_SIZE,a++)
  {  /* b is temporary: */
    dir=bf_buffer(dstart,&b,&good_entry,sizeof(good_entry));

    pdir.good_entry=(*dir->name!=0xe5&&*dir->name!=0);

        /* name mapping: */
    for(b=0;b<DOS_NAME_LEN&&dir->name[b];b++)
       map_name[b]=ascii_map[dir->name[b]];
    for(b=DOS_NAME_LEN-1;b>=0;b--)
    { if(map_name[b]==' ') map_name[b]='\0';
    }

        /* ext mapping: */
    for(b=0;b<DOS_EXT_LEN&&dir->ext[b];b++)
       map_ext[b]=ascii_map[dir->ext[b]];
    for(b=DOS_EXT_LEN-1;b>=0;b--)
    { if(map_ext[b]==' ') map_ext[b]='\0';
    }

    size=dir->sz4*(1<<24)+dir->sz3*(1<<16)+dir->sz2*(1<<8)+dir->sz1;
    cluster=dir->clust2*(1<<8)+dir->clust1;
    posn=cluster*dos_csize+dos_cstart;

    sprintf(name,"%.*s%s%.*s",DOS_NAME_LEN,map_name,*map_ext?".":""
       ,DOS_EXT_LEN,map_ext);

    if(*name==0) pdir.good_entry=0;
    pdir.start=dstart;
    pdir.name=name;
    pdir.attrib=dir->attrib;
    pdir.cluster=cluster;
    pdir.posn=posn;
    pdir.size=size;
    (*interpret)(&pdir);
  }
  *start=dstart;
}

static void print_dir(struct processed_dos_dir *dir)
{ printf("%8x %-*s  %2x  %4x %8x   %8x %8x\n",dir->start
     ,DOS_TOTAL_LEN,dir->name,dir->attrib
     ,dir->cluster,dir->size,dir->posn,dir->posn+dir->size);
}

void do_dos_dir(void)
{ position end,posn;
  static position dstart;
  int a;

  posn=dstart;
  get_range(&dstart,&end,DOSDIR_STD_SZ);

#ifdef INC_DOSDIR_HDR
#ifndef INC_DOSDIR_HDR_ALL
  if(posn!=dstart)
#endif /*INC_DOSDIR_HDR_ALL*/
  {/*      |xxxxxxxx nn-n  aa  cccc ssssssss   ssssssss eeeeeeee| */
    printf("    POSN %-*s ATR CLUST     SIZE      START      END\n"
       ,DOS_TOTAL_LEN,"NAME");
    for(a=0;a<48+DOS_TOTAL_LEN;a++) putchar('=');
    putchar('\n');
  }
#endif /*INC_DOSDIR_HDR*/

  do_scan_dir(&dstart,end,print_dir);
}

static char current_dos_dir[MAXSTRLEN+1];
static FILE *dos_cpio_file;

static void dos_recurse_handle(struct processed_dos_dir *dir)
{ char temp[MAXSTRLEN+1];
  int len;

  if(!dir->good_entry) return;

  if(!strcmp(dir->name,".")||!strcmp(dir->name,"..")) /* recursive linkages */
     return;

       /* lower case the name: */
  for(len=0;dir->name[len];len++) dir->name[len]=tolower(dir->name[len]);

  if(dir->attrib&0x10) /* subdirectory? */
  { len=strlen(current_dos_dir);
    sprintf(current_dos_dir+len,"%s/",dir->name);
    do_scan_dir(&dir->posn,dir->posn+dos_csize,dos_recurse_handle);
    current_dos_dir[len]='\0';
  }
  else /* normal file */
  { sprintf(temp,"%06o%06o%06o%06o%06o%06o%06o%06o%011o%06o%011o%s%s"
     /* magic dev inode       mode    u,g,ln,rdev,time */
       ,070707,0,dir->cluster,0100644,0,0,1 ,0   ,0
     /*  namesize                                   filesize */
       ,strlen(current_dos_dir)+strlen(dir->name)+1,dir->size
       ,current_dos_dir,dir->name);
    printf("%s\n",temp);
    if(fwrite(temp,1,strlen(temp)+1,dos_cpio_file)!=strlen(temp)+1) error("Error writing file");
    do_file_write(dir->posn,dir->posn+dir->size,dos_cpio_file);
  }
}

void dos_recurse(position start)
{ position end=start+dos_csize;
  char tempstr[MAXSTRLEN+1];
  int a;
  int left;

  l_getstr(tempstr);
  dos_cpio_file=fopen(tempstr,"w");
  if(!dos_cpio_file)
  { fprintf(stderr,"Unable to open file %s.\n",tempstr);
  }
  else
  { current_dos_dir[0]='\0';
    do_scan_dir(&start,end,dos_recurse_handle);
    sprintf(tempstr,"%06o%06o%06o%06o%06o%06o%06o%06o%011o%06o%011o%s"
   /*   magic dev inode  mode    u,g,ln,r,tm,namesize              */
       ,070707,0,0      ,0100644,0,0,1 ,0,0 ,strlen("TRAILER!!!")+1
       ,0 /* filesize */, "TRAILER!!!");
    if(fwrite(tempstr,1,strlen(tempstr)+1,dos_cpio_file)!=strlen(tempstr)+1)
      error("Error writing file");

         /* pad to a block */
    left=ftell(dos_cpio_file)&(512-1); /* actually; amount in last block */
    if(left)
    { for(a=0;a<MAXSTRLEN+1;a++) tempstr[a]='\0';
      left=512-left; /* adjust to whats really left */
    }
    while(left)
    { a=MAXSTRLEN+1;
      if(a>left) a=left;
      if(fwrite(tempstr,1,a,dos_cpio_file)!=a)
        error("Error writing file");
      left-=a;
    }

    fclose(dos_cpio_file);
  }
}

#endif /*INC_DOS_DIR*/
