/* bufr.c - I/O buffer code */

#include "general.h"


   /* medium size memory storage */

/* This does some not-quite legal type casting on stuff returned by
 * malloc(). */

#define M_FIRST_CHUNK   256
static byte **Mchunks;

byte *Mcpy(byte *buf,int len)
{ static byte *chunk;
  static int sz,lastsz;
  byte *ptr;

  len=(len+3)&NOT(3);

  if(!Mchunks)
  { sz=0;
    lastsz=M_FIRST_CHUNK;
  }

  if(len>sz)
  { sz=(lastsz=((lastsz+len)*3/2+7)&NOT(3))*sizeof(Mchunks)-sizeof(Mchunks);
    chunk=malloc(lastsz*sizeof(chunk)); /* not SO(*chunk) */
    if(!chunk) fatalerror("Out of memory in Mcpy()");
    *(byte ***)chunk=Mchunks;
    Mchunks=(byte **)chunk;
    chunk+=sizeof(Mchunks);
  }

  ptr=chunk;
  sz-=len;
  chunk+=len;

  memcpy(ptr,buf,len);
 return ptr;
}

void Mfreeall(void)
{ byte **next;

  for(;Mchunks;Mchunks=next)
  { next=(byte **)*Mchunks;
    free(Mchunks);
  }
}


   /* buffer management functions */

#ifdef TESTING_BUF
# define MAX_BF_LEN    16 /* MUST be a power of 2 for BASE_MASK to work */
#else /* ! TESTING_BUF */
# define MAX_BF_LEN   16384
#endif /*TESTING_BUF*/
#define BF_OFF_MASK   (MAX_BF_LEN-1) /* chunk_off=posn&OMASK */
#define BF_BASE_MASK  (NOT(BF_OFF_MASK)) /* chunk_base=posn&BMASK */
#define BF_CNT        16
#define BF_ENDLIST    NULL
#define BF_NOT_HERE   (-1)

static struct bf_chunk
{ position off;
  int len;
  int modified;
  struct bf_chunk *older,*newer;
  byte buf[MAX_BF_LEN];
} *bf;
static struct bf_chunk *bf_oldest;
static struct bf_chunk *bf_newest;

static FILE *bf_infofile; /*HERE*/
static position bf_read_posn,bf_write_posn;
static int bf_iswriteable; /* how file is currently fopen()ed */
position infosz,real_infosz;

position infosz,real_infosz;
char file_name[MAXSTRLEN+1];
int bf_loaded;
int bf_writeable;   /* my current state */

void bf_debug(void)
{ int cnt;

  for(cnt=0;cnt<BF_CNT;cnt++)
  { fprintf(stderr,"%x(%x) %x  o %x(%x)  n %x(%x)\n",bf+cnt,cnt,bf[cnt].off
       ,bf[cnt].older,bf[cnt].older-bf,bf[cnt].newer,bf[cnt].newer-bf);
  }
}

void bf_init(char *name,int loadfile)
{ int a;

  strcpy(file_name,name);
  if(bf_loaded) fclose(bf_infofile); /* close old file */ /*HERE*/
  if(loadfile)
  {   /* open file */
    bf_infofile=fopen(name,"r"); /*HERE*/
    bf_iswriteable=0;
    if(!bf_infofile)
    { if(bf_writeable)
      { bf_infofile=fopen(name,"w+"); /*HERE*/
        if(bf_infofile)
        { bf_iswriteable=1;
         goto isopen;
        }
      }
      fprintf(stderr,"Cannot open file \"%s\" for reading.\n",name);
      loadfile=0;
     goto dont_load;
    }
   isopen:

       /* find file length: */
    fseek(bf_infofile,0,2); /* seek to EOF */ /*HERE*/
    real_infosz=infosz=bf_read_posn=bf_write_posn=ftell(bf_infofile); /*HERE*/
    if(!bf_iswriteable)
    { if(bf_writeable==2)
      { bf_writeable=0;
        bf_set_writeable(2);
      }
      else bf_writeable=0;
    }
  }
  else
  {dont_load:
    bf_iswriteable=0;
    real_infosz=infosz=bf_read_posn=bf_write_posn=0;
  }
  bf_loaded=loadfile;

    /* get space */
  if(!bf)
  { bf=(struct bf_chunk *)malloc(BF_CNT*sizeof(*bf));
    if(!bf)
    { fprintf(stderr,"Unable to obtain a buffer\n");
     exit(-1);
    }
  }

    /* init buffers */
  for(a=0;a<BF_CNT;a++)
  { bf[a].off=BF_NOT_HERE;
    bf[a].older=bf+a-1;
    bf[a].newer=bf+a+1;
    bf[a].len=MAX_BF_LEN;
    bf[a].modified=0;
  }
    /* take care of list endpoints: */
  bf_oldest=bf;
  bf_newest=bf+BF_CNT-1;
  bf[0].older=BF_ENDLIST;
  bf[BF_CNT-1].newer=BF_ENDLIST;

  if(!bf_loaded)
  { bf[0].off=0;
    bf[0].len=0;
  }
}

static void bf_write_back(struct bf_chunk *ptr)
{ if(!ptr->modified) return;
  if(!bf_writeable||!bf_loaded)
  { fprintf(stderr
       ,"error: READONLY set or not loaded: loosing modifications %#x to %#x\n"
       ,ptr->off,ptr->off+ptr->len-1);
   return;
  }

  if(bf_write_posn!=ptr->off)
  { fseek(bf_infofile,ptr->off,0); /*HERE*/
  }

  bf_read_posn=BF_NOT_HERE;
  if(fwrite(ptr->buf,1,ptr->len,bf_infofile)!=ptr->len) /*HERE*/
  { fprintf(stderr,"write back didn't work right %#x to %#x\n"
       ,ptr->off,ptr->off+ptr->len-1);
  }
  ptr->modified=0;
  bf_write_posn=ptr->off+ptr->len;
  if(bf_write_posn>real_infosz) real_infosz=bf_write_posn;
}

void bf_write_back_all(void)
{ struct bf_chunk *ptr;
  int didcheck=0;

 redoit:
  for(ptr=bf_newest;ptr!=BF_ENDLIST;ptr=ptr->older)
  { if(ptr->modified) bf_write_back(ptr);
  }
  if(!didcheck&&real_infosz<infosz&&bf_writeable)
     /* this test applies if changesize; writeable 0; (flush); writeable 1 */
  { bf_loadchunk(real_infosz,&didcheck); /* didcheck used as dump */
    bf_modifiing(real_infosz);
    didcheck=1;
   goto redoit;
  }
}

static struct bf_chunk *bf_getchunk(position off)
   /* if chunk containing byte off is in cache, return it; else return
    * BF_ENDLIST.  If here, it also updates its access time. */
{ position base;
  struct bf_chunk *ptr;
#ifdef TESTING_BUF
  int cnt;
#endif /*TESTING_BUF*/

    /* check if here */
  base=off&BF_BASE_MASK;
#ifdef TESTING_BUF
  for(cnt=0,ptr=bf_newest;ptr!=BF_ENDLIST&&cnt<BF_CNT;ptr=ptr->older,cnt++)
#else /* ! TESTING_BUF */
  for(ptr=bf_newest;ptr!=BF_ENDLIST;ptr=ptr->older)
#endif /*TESTING_BUF*/
  { if(ptr->off==base)
    {   /* is here; need to move in list? */
      if(ptr->newer!=BF_ENDLIST)
      {   /* unlink its prev. position: */
        ptr->newer->older=ptr->older;
        if(ptr->older!=BF_ENDLIST) ptr->older->newer=ptr->newer;
        else bf_oldest=ptr->newer;

          /* link at newest */
        ptr->older=bf_newest;
        ptr->newer=BF_ENDLIST;
        bf_newest->newer=ptr;
        bf_newest=ptr;
      }
     return ptr;
    }
  }

#ifdef TESTING_BUF
    /* debug block: */
  if(ptr!=BF_ENDLIST) /* inconsistency */
  { bf_debug();
    for(cnt=0;cnt<10000000;cnt++) /*NOTHING*/ ;
  }
#endif /*TESTING_BUF*/

 return BF_ENDLIST;
}

void bf_set_writeable(int newval)
    /* set writeable flag's value.
     * 0 - readonly
     * 1 - writeable, but change to 0 when load new file.
     * 2 - continuous writeable.
     */
{ if(newval!=0&&newval!=1&&newval!=2)
    error("illegal value for writeable flag");
  if(newval==bf_writeable) return;

  bf_writeable=newval;

  if(!bf_loaded)
  { switch(newval)
    { case 1:
       error("Can't set nonpersistent writeable when no file loaded.");
      case 2:
      return;
      /* case 0: nothing */
    }
  }

  if(bf_writeable&&!bf_iswriteable)
  { bf_iswriteable=1;
    fclose(bf_infofile); /*HERE*/
    bf_infofile=fopen(file_name,"r+"); /*HERE*/
    if(!bf_infofile)
    { fprintf(stderr,"Unable to open file %s for update\n",file_name);
     exit(-1);
    }
    bf_write_posn=bf_read_posn=0;
  }
}

void bf_modifiing(position off)
    /* set cache's modified flag for byte off. */
{ position base;
  struct bf_chunk *ptr;

  base=off&BF_BASE_MASK;
  for(ptr=bf_oldest;ptr!=BF_ENDLIST;ptr=ptr->newer)
  { if(ptr->off==base)
    { if(!ptr->modified&&!bf_writeable)
      { fprintf(stderr,"Modifications %#x to %#x might be lost\n"
          ,ptr->off,ptr->off+ptr->len);
      }
      ptr->modified=1;
     return;
    }
  }
  fatalerror("internal: attempted set modified on non-loaded");
}

byte *bf_loadchunk(position off,int *size)
   /* return a pointer to byte off, and return total # of chars available
    * there in size.  Loads from disk if necessary.
    */
{ struct bf_chunk *ptr;
  position base;

    /* check if here */
  ptr=bf_getchunk(off);
  if(ptr==BF_ENDLIST)
  { if(!bf_loaded)
    { ptr=bf_oldest;
    }
    else
    {   /* write back oldest */
      if(bf_oldest->modified)
      { bf_write_back(bf_oldest);
      }

        /* unlink oldest, link newest */
      ptr=bf_oldest;
      bf_oldest=ptr->newer;
      ptr->newer->older=BF_ENDLIST;
      ptr->newer=BF_ENDLIST;
      bf_newest->newer=ptr;
      ptr->older=bf_newest;
      bf_newest=ptr;

        /* fill in */
      base=off&BF_BASE_MASK;
      ptr->off=base;
      if(bf_read_posn!=base)
      { fseek(bf_infofile,base,0); /*HERE*/
      }
      bf_write_posn=BF_NOT_HERE;
      ptr->len=fread(ptr->buf,1,MAX_BF_LEN,bf_infofile); /* HERE*/
      if(ptr->len==-1&&base==infosz) ptr->len=0; /* compiler bug work around */
      if(ptr->len!=MAX_BF_LEN&&base+ptr->len!=infosz)
      { if(base+ptr->len==real_infosz)
        { while(ptr->len<MAX_BF_LEN&&base+ptr->len<infosz)
          { ptr->buf[ptr->len++]=0;
          }
          if(ptr->len>=MAX_BF_LEN&&base+ptr->len<infosz)
             fprintf(stderr,"Internal file length error reading file\n");
        }
        else fprintf(stderr,"Error reading input file stat %d\n",ptr->len);
      }
      ptr->modified=0;
      bf_read_posn=ptr->off+ptr->len;
    }
  }

  *size=ptr->len-(off&BF_OFF_MASK);
 return ptr->buf+(off&BF_OFF_MASK);
}

byte *bf_loadchunk_back(position off,int *size)
{ position realoff,partsize;
  byte *ptr;

  realoff=(off-1)&BF_BASE_MASK;
  ptr=bf_loadchunk(realoff,&partsize);

  *size=off-realoff;
  if(partsize<*size)
     error("Internal error: bf_loadchunk_back() size consistency");

 return ptr;
}

int bf_changesize(position newsz)
    /* change file size (doesn't properly reduce size if its requested) */
{ byte *ptr;
  struct bf_chunk *bfptr;
  int size;
  int a;

  if(!bf_writeable||!bf_loaded)
  { fprintf(stderr,"Can't change size when readonly or not loaded\n");
   return 0;
  }
  if(newsz<infosz)
  { fprintf(stderr,"Reduction of size will not be reflected in real file.\n");
       /* free up cache in higher posns: */
    for(bfptr=bf_newest;bfptr!=BF_ENDLIST;bfptr=bfptr->older)
    { if(bfptr->off>=newsz)
      { bfptr->modified=0;
        bfptr->off=BF_NOT_HERE;
      }
    }
/*HEREfuture: when fixed, write real_infosz properly */
  }
    /* allocate whole new chunks: */
  if((infosz&BF_BASE_MASK)<(newsz&BF_BASE_MASK))
  { ptr=bf_loadchunk(infosz,&size);
    bfptr=bf_getchunk(infosz);
    if(!bfptr) fatalerror("Lost buffer chunk");

      /* first partial chunk: */
    size=MAX_BF_LEN-(infosz&BF_OFF_MASK);
    if(size>newsz-infosz) size=newsz-infosz;
    infosz+=size;
    bfptr->len+=size;
    while(size--) *(ptr++)=0;

    bfptr->modified=1;
    bf_write_back(bfptr);
    if(infosz<newsz) for(a=0;a<MAX_BF_LEN;a++) bfptr->buf[a]=0;

    while(infosz<(newsz&BF_BASE_MASK)) /* whole chunks */
    { bfptr->off+=MAX_BF_LEN;
      bfptr->modified=1;
      bf_write_back(bfptr);
      infosz+=MAX_BF_LEN;
    }
    if(infosz<newsz) /* last chunk */
    { bfptr->off+=MAX_BF_LEN;
      bf_modifiing(infosz);
    }
  }
  else if(infosz<newsz) /* extend this chunk only */
  { ptr=bf_loadchunk(infosz,&size);
    bf_modifiing(infosz);
    size=newsz-infosz;
    while(size--) *(ptr++)=0;
    bfptr=bf_getchunk(newsz);
    if(!bfptr) fatalerror("internal; missing chunk");
  }
  bfptr->len=newsz-bfptr->off;
  infosz=newsz;
 return 1;
}

byte *bf_16(position off,int *size)
     /* get 16 byte chunk in static buffer (buffer is shared between calls,
      * so if you need to keep it around, make copy after calling this.)
      */
{ int size1,size2;
  byte *ptr;
  static byte int_buf[16];

  *size=16;
  ptr=bf_loadchunk(off,&size1);
  if(size1<16)
  { memcpy(int_buf,ptr,size1);
    ptr=bf_loadchunk(off+size1,&size2);
    if(size1+size2<16)
    { memcpy(int_buf+size1,ptr,size2);
      *size=size1+size2;
    }
    else
    { memcpy(int_buf+size1,ptr,16-size1);
    }
   return int_buf;
  }
 return ptr;
}

void *bf_buffer(position off,int *size,void *buf,int rsize)
     /* get rsize chunk from off.  give directly if possible; else copy
      * into buf and return copy. return actual size gotten in size. */
{ int size1,size2;
  byte *ptr;

  *size=rsize;
  ptr=bf_loadchunk(off,&size1);
  if(size1<rsize)
  { memcpy(buf,ptr,size1);
    while(1)
    { ptr=bf_loadchunk(off+size1,&size2);
      if(size2==0||size2==-1)
      { *size=size1;
       break;                /* exit 1, can't completely fill. */
      }
      else if(size1+size2<rsize)
      { memcpy((byte *)buf+size1,ptr,size2);
        size1+=size2;
      }
      else
      { memcpy((byte *)buf+size1,ptr,rsize-size1);
       break;                 /* exit 2, filled nicely; normal exit */
      }
    }
   return buf;
  }
 return ptr;
}

byte bf_byte(position off)
{ int size;
  byte *ptr;

  ptr=bf_loadchunk(off,&size);
 return *ptr;
}

word bf_word(position off)
{ int size;
  byte *ptr;

  ptr=bf_loadchunk(off,&size);
  if(size<2)
  { ptr=bf_16(off,&size);
    if(size<2) return 0;
  }
 return *(word *)ptr;
}

dword bf_dword(position off)
{ int size;
  byte *ptr;

  ptr=bf_loadchunk(off,&size);
  if(size<4)
  { ptr=bf_16(off,&size);
    if(size<4) return 0;
  }
 return *(dword *)ptr;
}

#ifdef BAD_COMPILER
int memcmp(byte *s1,byte *s2,position len)
{ while(len--) if(*s1++!=*s2++) return s1[-1]-s2[-1];
 return 0;
}
#endif /*BAD_COMPILER*/

int bf_memcmp(position off,byte *s2,position len)
   /* compare file position off with memory at s2. */
{ byte *ptr;
  int size;
  int i;

  ptr=bf_loadchunk(off,&size);
  while(size<len)
  { i=memcmp(ptr,s2,size);
    if(i) return i;
    len-=size;
    off+=size;
    s2+=size;
    ptr=bf_loadchunk(off,&size);
  }
 return memcmp(ptr,s2,len);
}
